/**
* client.c - P2PWNC client
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "io.h"
#include "crypt.h"
#include "client_rdb.h"
#include "conffile.h"
#include <errno.h>

#include <sys/ioctl.h>
#include <net/if.h>

#define CLIENT_STATE_OFF 10
#define CLIENT_STATE_PENDING 20
#define CLIENT_STATE_ON 30

//default configuration file paths...
#define DEF_CONF_PATH "../conf/p2pwnc_client.conf"
#define DEF_CONF_FNAME "p2pwnc-client.conf"

unsigned short client_state = CLIENT_STATE_OFF;
PWNC_CERT *client_cert;
PWNC_PUBKEY *client_pub;
PWNC_PUBKEY *server_pub;
PWNC_PRKEY *client_priv;
unsigned long timestamp;
struct sockaddr_in server_address;
int sockfd = 0;
int update_mode = 0; //update_mode = 1 ==> update receipt repository
FILE *conf_file_fp = NULL;

//name of network interface
char * ifname = NULL;
char def_ifname[32];

//initial network interface traffic
//(that is before the session has begun)
long initial_traffic = 0L;

//tolerance in client/ap traffic measurements (in bytes)
int tolerance = 0;

void cleanup_session() {
	SAFE_FREE_KEY(client_pub);
	SAFE_FREE_KEY(client_priv);
	SAFE_FREE_CERT(client_cert);
}

//reads public and private key from disk
int read_keys(const char *pub_key_path, const char *priv_key_path, const char *cert_path) {
	client_pub = read_key_file(pub_key_path);
	if (!client_pub) {
		return 0;
	}
	client_priv = read_key_file(priv_key_path);
	if (!client_priv) {
		SAFE_FREE_KEY(client_pub);
		return 0;
	}
	client_cert = read_cert_file(cert_path);
	if (!client_cert) {
		SAFE_FREE_KEY(client_pub);
		SAFE_FREE_KEY(client_priv);
		return 0;
	}
	return 1;
}

int connect_to_server(const char *srv_ip_addr, int portnum) {
	int addr_len = sizeof(server_address);
	int i = 0;

	memset((char  *)&server_address, 0, sizeof(server_address));
	server_address.sin_family = AF_INET;
	if (!inet_aton(srv_ip_addr, &(server_address.sin_addr))) {
		return 0;
	}
	//server_address.sin_port = htons(PWNC_SERVER_PORT);
	server_address.sin_port = htons(portnum);

	sockfd = socket (PF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) {
		fprintf(stderr, "connect_to_server: socket creation failed\n");
		return 0;
	}

	for (i=0; i<3; i++) {
		fprintf(stderr, "connect_to_server: Connecting to server...\n");
		if (connect(sockfd, (const struct sockaddr *)&server_address, addr_len) == 0) {
			return 1;
		}
	}
	fprintf(stderr, "connect_to_server: Connection failed\n");
	return 0;
}

#ifdef OFFER_RECEIPTS
typedef struct cli_rbtree_range_query_func_data_t {
	int sockfd;
} RQDATA_CL;

//for each receipt, this func will be called...
//each receipt will be sent to the client in an RCPT message
void client_range_query_callback(RBTNODE *node, void *data) {
	unsigned char *text = NULL; //text message to be sent

	text = receipt_2_text( ((RBTREE_RCPT_CL*)node->data)->rcpt );
	if (!text) {
		//error constructing rcpt message
		return;
	}

	//write message to socket
	if (!write( ((RQDATA_CL*)data)->sockfd, text, strlen(text))) {
		fprintf(stderr, "range_query: Error writing to socket\n");
	}

	//free message text
	SAFE_FREE(text);
}
#endif

int init_session() {
	int server_mode = 0;

#ifdef OFFER_RECEIPTS
	//client will start the session offering receipts
	RQDATA_CL *rqd = (RQDATA_CL*)malloc(sizeof(RQDATA_CL));
	unsigned long timestamp = 0; //give me all the receipts
	memset(rqd, 0, sizeof(RQDATA_CL));
	rqd->sockfd = sockfd;

	//range query (retrieve receipts from database and send them to server)
	client_rdb_range_query( client_range_query_callback, (void*)rqd, timestamp );
	SAFE_FREE(rqd);
	//now continue session normally
#endif

	unsigned char *text = construct_conn_msg(client_cert);
	int mtype = 0;
	if (!text) {
		fprintf(stderr, "Error constructing conn msg\n");
		return 0;
	}
	if (!write(sockfd, text, strlen(text))) {
		fprintf(stderr, "Error writing to socket\n");
		SAFE_FREE(text);
		return 0;
	}
	else {
		text = read_message(sockfd, &mtype);
		fprintf(stderr, "%s\n\n", text);
	}
	if (mtype != PWNC_MSG_CACK) {
		fprintf(stderr, "Error receiving CACK message\n");
		SAFE_FREE(text);
		return 0;
	}
	else {
		if (!(server_mode = parse_cack_msg(text, &timestamp))) {
			fprintf(stderr, "Error parsing CACK msg\n");
			SAFE_FREE(text);
			return 0;
		}
		//now check if the server needs to be sent receipts to admit me
		if (server_mode == MODE_RCPT_REQ_ON) {
			//TODO: SEND RECEIPTS TO TEAM SERVER
			//needs to find receipts somehow ----> client rdb (prepei na pros8esw ki edw ena #define _TEAM_SERVER mallon....TODO)
			SAFE_FREE(text);
			return 0;
		}
		else
		if (server_mode == MODE_RCPT_REQ_OFF) {
			SAFE_FREE(text);
			client_state = CLIENT_STATE_ON;
			return 1;
		}
		else {
			SAFE_FREE(text);
			//client_state = CLIENT_STATE_OFF??;
			return 0;
		}
	}
}

//gets the interface name associated with the client's ip address
//to do so, ioctl on socket
char * get_interface_name() {
	struct ifreq ir;
	int len = 0;
	struct sockaddr_in sock_name;
	struct sockaddr_in if_addr;

	unsigned char *interface_name = NULL;

	memset(&ir, 0, sizeof(struct ifreq));

	interface_name = (unsigned char *)malloc(32);
	memset(interface_name, 0, 32);

	memset(&sock_name, 0, sizeof(struct sockaddr_in));
	len = sizeof(struct sockaddr_in);
	if (getsockname(sockfd, (struct sockaddr*)&sock_name, &len) == -1) {
		SAFE_FREE(interface_name);
		return NULL;
	}


	int i = 0; //interface index
	while(1) {
		i++;
		ir.ifr_ifindex = i;
		//get interface name given the interface index
		if (ioctl(sockfd, SIOCGIFNAME, &ir) == -1) {
			//if this ioctl fails, the interface index has exceeded the number
			//of interfaces on the machine, so --> break
			SAFE_FREE(interface_name);
			printf(strerror(errno));
			break;
		}

		//for the specified interface name, retrieve the IP address
		//associated with it
		if (ioctl(sockfd, SIOCGIFADDR, &ir) == -1) {
			//if it fails, do not break but continue loopin
			//(until the other ioctl fails)
			SAFE_FREE(interface_name);
			continue;
		}

		memset(&if_addr, 0, sizeof(struct sockaddr_in));
		memcpy(&if_addr, &ir.ifr_addr, sizeof(struct sockaddr_in));

		//compare ip address of socket to ip address of interface
		//if they match, this is the interface name we're seeking for
		if ( if_addr.sin_addr.s_addr == sock_name.sin_addr.s_addr) {
			strcpy(interface_name, ir.ifr_name);
			break;
		}
	}

	return interface_name;
}

//returns measured traffic
unsigned long get_traffic() {
	//read /proc/net/dev info
	FILE * fp = NULL;

	fp = fopen("/proc/net/dev", "r");
	if (!fp) {
		fprintf(stderr, "get_traffic: Could not open /proc/net/dev\n");
		return -1;
	}

	long bytes_in = 0L;
	long bytes_out = 0L;

	int linelen = 1024;
	char * line = NULL;

	line = (char*)malloc(linelen);

	//skip the first 2 lines of /proc/net/dev
	memset(line, 0, linelen);
	getline(&line, &linelen, fp);
	memset(line, 0, linelen);
	getline(&line, &linelen, fp);

	char metrix[1024];
	memset(line, 0, linelen);
	memset(metrix, 0, linelen);


	int i = 0;
	while(getline(&line, &linelen, fp) != -1) {
		if (strstr(line, ifname)) {
			strcpy(metrix, strstr(line, ":") + 1);
			int j = 0;
			while(metrix[j] == ' ') {
				j++;
			}
			sscanf( (metrix+j), "%lu %*lu %*lu %*lu %*lu %*lu %*lu %*lu %lu %*lu %*lu %*lu %*lu %*lu %*lu %*lu", &bytes_in, &bytes_out );
			break;
		}
		memset(line, 0, linelen);
		memset(metrix, 0, linelen);
	}

	SAFE_FREE(line);
	fclose(fp);

	return (bytes_in + bytes_out);
}

#ifdef _TEAM_SERVER
extern RBTNODE* NULL_NODE;
extern int num_of_receipts_in_tree;

//handles a receipt received by the team server during an UPDT session
//the receipt is stored into the receipts database
//text: RCPT text message
void handle_receipt(unsigned char *text) {
	PWNC_RCPT *rcpt = NULL; //latest rcpt
	rcpt = text_2_rcpt(text); //parse message
	if (!rcpt) {
		return;
	}
	if (insert_receipt(rcpt) == -1) {
		//error...resources must be freed
		SAFE_FREE_RCPT(rcpt);
	}
}

int update_receipt_db() {
	unsigned long most_recent_ts = 0; //timestamp of the most recent receipt
	unsigned char *text = NULL; //an UPDT text message
	unsigned char *text_data = NULL; //data of an UPDT message
	int mlen = 0;
	int mtype = 0;

	//find most recent receipt in database
	most_recent_ts = get_most_recent_ts();

	//send updt message to server to get recent receipts
	text_data = make_updt_req_data(most_recent_ts, &mlen);
	if (!text_data) {
		return -1;
	}
	text = construct_updt_msg(text_data, mlen);
	fprintf(stderr, "%s\n", text);
	if (!text) {
		SAFE_FREE(text_data);
		return -1;
	}
	if (!write(sockfd, text, strlen(text))) {
		SAFE_FREE(text);
		SAFE_FREE(text_data);
		fprintf(stderr, "Error writing to socket\n");
		return -1;
	}

	while(1) {
		//TODO: SELECT on socket...(to check timeouts)
		//for each receipt received-->store_receipt
		SAFE_FREE(text);
		text = read_message(sockfd, &mtype);
		if (!text) {
			return -1;
		}
		if (mtype == PWNC_MSG_UPDT) {
			//the only acceptable updt message at this poing is: UPDT\r\nContent-length: 3\r\nEND --->means end of receipts
			if (!strncasecmp(text, "UPDT\r\nContent-length: 3\r\nEND", strlen("UPDT\r\nContent-length: 3\r\nEND"))) {
				SAFE_FREE(text);
				break;
			}//else ignore message
		}
		else
		if (mtype == PWNC_MSG_RCPT) {
			//normal case (receipt from team server)
			handle_receipt(text);
		}
	}
	return 0;
}
#endif


void print_usage() {
	fprintf(stderr, "\nUsage: client ");
	fprintf(stderr, "[-h/--help] ");
	fprintf(stderr, "[-c/--conf file] ");
	fprintf(stderr, "[-pk/--public-key-path file] ");
	fprintf(stderr, "[-sk/--secret-key-path file] ");
	fprintf(stderr, "[-crt/--cert-path file] ");
	fprintf(stderr, "[-srv/--server-addr ip_address] ");
	fprintf(stderr, "[-p/--server-port port] ");
	fprintf(stderr, "[-t/--tolerance bytes] ");
	fprintf(stderr, "[-u/--update] \n\n");
	fprintf(stderr, "\t-h --help\t\tDisplay this message\n");
	fprintf(stderr, "\t-c --conf\t\tClient configuration file\n");
	fprintf(stderr, "\t-pk --public-key-path\tClient public key file\n");
	fprintf(stderr, "\t-sk --secret-key-path\tClient secret(private) key file\n");
	fprintf(stderr, "\t-crt --cert-path\tClient certificate file\n");
	fprintf(stderr, "\t-srv --server-addr\tServer IP address. If -u is given, this is the team server address. ");
	fprintf(stderr, "Otherwise, it is the IP address of the access point to connect\n");
	fprintf(stderr, "\t-t --tolerance\t\tMaximum difference (in bytes) between the access point and the client measured traffic in an RREQ message\n");
	fprintf(stderr, "\t-u --update\t\tUpdate mode: communicate with the team server and retrieve receipt repository\n");
	fprintf(stderr, "\t-p --server-port\tServer port\n");
	fprintf(stderr, "\t-if --ifname\tNetwork interface (e.g. wlan0)\n\n");

}

int main (int argc, char **argv) {
	//client-computed traffic
	long client_traffic = 0;
	//ap-reported traffic
	long traffic = 0;
	//message type
	int mtype = 0;
	//message text
	unsigned char *text = NULL;
	//receipt constructed and sent to the ap
	PWNC_RCPT *rcpt = NULL;

	//seed pseudo number generator
	init_prng();

//##############################################
	//configuration
	char str_pub_key_path[80];
	char str_priv_key_path[80];
	char str_cert_path[80];
	char str_conf_file_path[80];
	char str_server_addr[20];
	int server_port = 0;
	int help = 0;

	memset(&str_pub_key_path, 0, 80);
	memset(&str_priv_key_path, 0, 80);
	memset(&str_cert_path, 0, 80);
	memset(&str_conf_file_path, 0, 80);
	memset(&str_server_addr, 0, 20);
	memset(&def_ifname, 0, 32);

//	memset(&str_port, 0, 10);

	CMD_ARG *cmd_line_args = NULL;
	CENTRY *conf_entries = NULL;
	unsigned char *valid_conf_names[] = {
		"public-key-path",
		"secret-key-path",
		"cert-path",
		"server-addr",
		"server-port",
		"ifname",
		"tolerance"
		};
	unsigned char *valid_cmd_line_names[] ={
		"conf",
		"public-key-path",
		"secret-key-path",
		"cert-path",
		"server-addr",
		"server-port",
		"ifname",
		"update",
		"help",
		"tolerance"
		};

	//c: conffile
	//pk: pubkey
	//sk: secretkey
	//crt:cert
	//srv: server address
	//p: port
	//u: update mode
	unsigned char* cmd_line_switches[] = {"c", "pk", "sk", "crt", "srv", "p", "if", "u", "h", "t"};

	unsigned short conf_var_types[] = {
		VTYPE_STRING, //pubkey
		VTYPE_STRING, //secret key
		VTYPE_STRING, //cert
		VTYPE_STRING, //server addr
		VTYPE_INT, //server port
		VTYPE_STRING, //ifname
		VTYPE_INT, //tolerance
	};
	unsigned short cmd_line_var_types[] = {
		VTYPE_STRING, //conf
		VTYPE_STRING, //pk
		VTYPE_STRING, //sk
		VTYPE_STRING, //crt
		VTYPE_STRING, //srv
		VTYPE_INT, //port
		VTYPE_STRING, //ifname
		VTYPE_NOVAL, //update
		VTYPE_NOVAL, //help
		VTYPE_INT //tolerance
	};

	void *conf_variables[] = {
		&str_pub_key_path,
		&str_priv_key_path,
		&str_cert_path,
		&str_server_addr,
		&server_port,
		&def_ifname,
		&tolerance
	};

	void *cmd_line_variables[] = {
		&str_conf_file_path,
		&str_pub_key_path,
		&str_priv_key_path,
		&str_cert_path,
		&str_server_addr,
		&server_port,
		&def_ifname,
		&update_mode,
		&help,
		&tolerance
	};

	conf_entries = make_valid_entries(valid_conf_names, conf_variables, conf_var_types, 7);
	cmd_line_args = make_valid_args(valid_cmd_line_names, cmd_line_switches, cmd_line_variables, cmd_line_var_types, 10);

	if (argc < 2) {
		//no command line flags specified, so parse default conf file
		//also, the -u option was not given
		conf_file_fp = fopen(DEF_CONF_PATH, "r");
		if (!conf_file_fp) {
			//first check in /etc/pwnc.conf
			conf_file_fp = fopen(DEF_CONF_FNAME, "r");
			if (!conf_file_fp) {
				//if fopen failed, try to open pwnc.conf in the current path
				//if all else fails ---> exit
				fprintf(stderr, "client: Error opening configuration file\n\n");
				print_usage();
				exit(1);
			}
		}
		//default conffile opened ok, so parse it
		parse_conf_file(conf_entries, 7, conf_file_fp);
	}
	else {
		//parse command line
		//if there is a -c flag, ignore all other flags (except for the -u flag)
		//and check the specified conf file
		parse_cmd_line_args(argc, argv, cmd_line_args, 10);
		if (help) {
			//user gave the -h flag (probably among others...)
			print_usage();
			exit(0);
		}
		if (strlen(str_conf_file_path)) {
			//there was a -c flag, so ignore all other flags (except -u) and check this conf file
			conf_file_fp = fopen(str_conf_file_path, "r");
			if (!conf_file_fp) {
				fprintf(stderr, "client: Warning: Failed to open %s, trying defaults\n", str_conf_file_path);
				//try to open specified conffile - if it fails, try defaults again
				conf_file_fp = fopen(DEF_CONF_PATH, "r");
				if (!conf_file_fp) {
					//trying default conffile path
					conf_file_fp = fopen(DEF_CONF_FNAME, "r");
					if (!conf_file_fp) {
						//trying to open default conf file in current dir
						fprintf(stderr, "client: Error opening configuration file\n");
						exit(1);
					}
				}
			}
			parse_conf_file(conf_entries, 7, conf_file_fp);
		}
		//otherwise: we are using cmd line arguments
	}

	//check mandatory fields (keys, mq key, ipt path)
	if (
		!strlen(str_pub_key_path) ||
		!strlen(str_priv_key_path) ||
		!strlen(str_cert_path) ||
		!strlen(str_server_addr) ||
		!server_port
	) {
		fprintf(stderr, "client: Some configuration information is missing\n\n");
		print_usage();
		exit(1);
	}

	fprintf(stderr, "Client configuration:\n");
	fprintf(stderr, "---------------------\n");
	fprintf(stderr, "Public key file: %s\n", str_pub_key_path);
	fprintf(stderr, "Secret key file: %s\n", str_priv_key_path);
	fprintf(stderr, "Client certificate file: %s\n", str_cert_path);
	fprintf(stderr, "Server address: %s\n", str_server_addr);
	fprintf(stderr, "Server port: %d\n", server_port);
	fprintf(stderr, "Network interface: %s\n", def_ifname);
	fprintf(stderr, "Tolerance: %d\n", tolerance);
	fprintf(stderr, "Updating repository: %s\n\n", (update_mode?"yes":"no") );

//##########################################################

	if (!read_keys(str_pub_key_path, str_priv_key_path, str_cert_path)) {
		fprintf(stderr, "client: Could not load keys and certificate. Check the given path\n");
		exit(1);
	}

	if (!connect_to_server(str_server_addr, server_port) ) {
		SAFE_FREE_KEY(client_pub);
		SAFE_FREE_KEY(client_priv);
		SAFE_FREE_CERT(client_cert);
		exit(1);
	}

#ifdef _TEAM_SERVER
	//if it is compiled with team server support
	//we have to init client rdb and set up UPDT capability
	tree_init();
	NULL_NODE->color = NODE_COLOR_BLACK;
	num_of_receipts_in_tree = 0;
	rcpt_tree = NULL_NODE;
	//first load data from "disk"
	load_receipts();

	if (update_mode) {
		//connect to server and update receipt repository
		update_receipt_db();
		return 0;
	}
#endif

	//get interface name
	if (!strlen(def_ifname)) {
		ifname = get_interface_name();
		if (!ifname) {
			fprintf(stderr, "Error: Could not retrieve interface name\n");
			cleanup_session();
			exit(1);
		}
	}
	else {
		ifname = (char*)malloc(32);
		memset(ifname, 0, 32);
		strncpy(ifname, def_ifname, strlen(def_ifname));
	}

	fprintf(stderr, "Interface name: %s\n", ifname);

	//retrieve initial traffic
	//(it will be subtracted from future measured values)
	initial_traffic = get_traffic();
	fprintf(stderr, "Initial traffic: %ld\n", initial_traffic);

	//if we are not updating...
	if (!init_session()) {
		SAFE_FREE_KEY(client_pub);
		SAFE_FREE_KEY(client_priv);
		SAFE_FREE_CERT(client_cert);
		fprintf(stderr, "client: Could not connect to server %s on port %d\n", str_server_addr, server_port);
		exit(1);
	}

	//now we're ok...
	//client waits for server messages and answers them
	while (1) {
		text = read_message(sockfd, &mtype);
		fprintf(stderr, "%s\n\n", text);
		if (mtype == PWNC_MSG_RREQ) {
			//parse rcpt request - get server public key
			server_pub = parse_rreq(text, &traffic);

			if (!server_pub) {
				fprintf(stderr, "Error extracting server public key\n");
				SAFE_FREE(text);
				cleanup_session();
				return 0;
			}
			SAFE_FREE(text);

			//TRAFFIC CHECKING
			//If ap-reported traffic exceeds client-measured traffic (+tolerance)
			//the session ends
			client_traffic = get_traffic() - initial_traffic;

			//fprintf(stderr, "Measured traffic: %ld\nInitial traffic: %ld\nClient Traffic: %ld\nAp Traffic:%ld\n", get_traffic(), initial_traffic, client_traffic, traffic);

			if ( (traffic - client_traffic) > tolerance ) {
				fprintf(stderr, "client: Tolerance exceeded. Ending session...\n");
				SAFE_FREE(text);
				SAFE_FREE_KEY(server_pub);
				cleanup_session();
				exit(1);
			}

			rcpt = construct_receipt(client_cert, server_pub, timestamp, traffic, client_priv, 0);
			if (!rcpt) {
				fprintf(stderr, "client: Error constructing receipt\n");
				SAFE_FREE(text);
				SAFE_FREE_KEY(server_pub);
				cleanup_session();
				return 0;
			}
			//send receipt
			text = receipt_2_text(rcpt);
			if (!text) {
				fprintf(stderr, "client: Error constructing RCPT msg\n");
				SAFE_FREE_RCPT(rcpt);
				cleanup_session();
				return 0;
			}
			if (!write(sockfd, text, strlen(text))) {
				SAFE_FREE_RCPT(rcpt);
				SAFE_FREE(text);
				cleanup_session();
				fprintf(stderr, "client: Error writing to socket\n");
				return 0;
			}

			fprintf(stderr, "client: sent receipt to server\n");
			fprintf(stderr, "%s\n", text);
			SAFE_FREE(text);
			SAFE_FREE_KEY(server_pub);
			SAFE_FREE_KEY(rcpt->RcptSig);
			SAFE_FREE(rcpt);
		}
	}
	close (sockfd);

	cleanup_session();
	return 0;
}
