/**
* client_rdb.h - Receipt storage at the client side
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*
Receipt storage at the client side

This module is responsible for communicating with the Team Server and updating the client's receipt repository
Receipts can be stored in the permanent storage of the client (e.g. disk, flash card, etc)
The client sends an UPDT msg (if operating in decentralized mode) to the srv and stores the received receipts

The receipts are stored in a red-black tree in memory.
The clients are expected to be small, memory constrained devices, so the tree has a limited number of nodes
If the num of receipts received by the team server exceed this limit, the "newest" are kept (replace the oldest ones)
*/

#ifndef _PWNC_CLIENT_RDB_H
#define _PWNC_CLIENT_RDB_H

#include "io.h"
#include "rbtree.h"

#ifndef RECEIPT_THRESHOLD
#define RECEIPT_THRESHOLD 5000
#endif

#ifndef RECEIPTS_DB_FILE
#define RECEIPTS_DB_FILE "../data/client_receipts.db"
#endif

#ifndef RECEIPTS_DB_FREE_LIST
#define RECEIPTS_DB_FREE_LIST "../data/client_receipts.fl"
#endif


#define RDB_ACTION_INSERT 10
#define RDB_ACTION_REMOVE 20

//this struct represents the "data" field of an RBTNODE
//it stores a receipt and its position in the file (permanent storage)
typedef struct _client_rbtree_rcpt_ {
	PWNC_RCPT *rcpt;
	int disk_pos;
} RBTREE_RCPT_CL;

RBTNODE *rcpt_tree;
int num_of_receipts_in_tree;

//RB tree CALLBACKS
//returns the key for an RBTNODE storing receipts (node->data is a PWNC_RCPT*)
void *rcpt_key_func(RBTNODE *node);
//sets key value to an RBTNODE storing receipts
void rcpt_set_key(RBTNODE *node, void *keyval);
//compares (the keys of) two RBTNODEs storing receipts
int rcpt_comp(RBTNODE *node1, RBTNODE *node2);

//constructs a new RBTNODE and returns it
//the new node contains the rcpt in the data field (that is an RBTREE_RCPT object containing rcpt)
RBTNODE *RBTNODE_new(PWNC_RCPT *rcpt, const void *key_func, const void *set_key_func, const void *comp_func);

//Opens the receipt db and loads the receipts in memory
void load_receipts();

//returns the timestamp of the most recent receipt
unsigned long get_most_recent_ts();

//adds/removes a tree node from disk
int handle_receipt_storage(RBTREE_RCPT_CL *to_be_handled, int action);
//write a receipt to disk and insert it into the tree
int insert_receipt(PWNC_RCPT *rcpt);

//range query function (calls rbtree's range_query)
void client_rdb_range_query( void (*do_stuff)(RBTNODE*, void *), void *do_stuff_args, unsigned long timestamp );

#endif /* _PWNC_CLIENT_RDB_H */
