/**
* rbtree.c - Red-Black tree data structure implementation
* (internal <--> network (ascii) representation, message generation/parsing, etc)
*
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/**
RED-BLACK tree data structure implementation
see CLRS chapter 13
Each tree node holds:
- Color (r/b)
- Receipt (node key: receipt timestamp) list (receipts with equal timestamp)
- left/right/parent pointers
*/

#include "rbtree.h"

//tree initialization - mainly alloc/set NULL_NODE
void tree_init() {
	NULL_NODE = (RBTNODE*)malloc(sizeof(RBTNODE));
	memset(NULL_NODE, 0, sizeof(RBTNODE));
	NULL_NODE->color = NODE_COLOR_BLACK;
	NULL_NODE->data = NULL;
	NULL_NODE->left = NULL_NODE;
	NULL_NODE->right = NULL_NODE;
	NULL_NODE->parent = NULL_NODE;
	NULL_NODE->key_func = NULL;
	NULL_NODE->set_key_func = NULL;
	NULL_NODE->comp_func = NULL;
}

//left rotation: swap x and its right child ==> y is the parent, x is the left child
void left_rotate(RBTNODE **tree, RBTNODE *x) {
	//set y to x's right child
	RBTNODE *y = x->right;
	if (x->right == NULL_NODE) {
		return;
	}

	x->right = y->left;
	if (y->left != NULL_NODE) {
		y->left->parent = x;
	}
	y->parent = x->parent;
	if (x->parent == NULL_NODE) {
		*tree = y;
	}
	else
	if (x == x->parent->left) {
		x->parent->left = y;
	}
	else {
		x->parent->right = y;
	}
	y->left = x;
	x->parent = y;
}

//right rotation: swap y(parent) and its left child ==> y becomes right child of x
void right_rotate(RBTNODE **tree, RBTNODE *y) {
	RBTNODE *x = y->left;
	if (y->left == NULL_NODE) {
		return;
	}

	y->left = x->right;
	if (x->right != NULL_NODE) {
		x->right->parent = y;
	}
	x->parent = y->parent;
	if (y->parent == NULL_NODE) {
		*tree = x;
	}
	else
	if (y == y->parent->right) {
		y->parent->right = x;
	}
	else {
		y->parent->left = x;
	}
	x->right = y;
	y->parent = x;
}


void rbtree_delete_fixup(RBTNODE **tree, RBTNODE *x) {
	RBTNODE *w = NULL_NODE;
	//printf("delete fixup\n");
	while ( (x != *tree) && (x->color == NODE_COLOR_BLACK) ) {
		//loop ends when we have ascended to the root OR when x has been colored black
		if (x == x->parent->left) {
			//if x is a left child
			w = x->parent->right;
			if (w->color == NODE_COLOR_RED) {
				w->color = NODE_COLOR_BLACK;
				x->parent->color = NODE_COLOR_RED;
				left_rotate(tree, x->parent);
				w = x->parent->right;
			}
			if ( (w->left->color == NODE_COLOR_BLACK) && (w->right->color == NODE_COLOR_BLACK) ) {
				w->color = NODE_COLOR_RED;
				x = x->parent;
			}
			else {
				if (w->right->color == NODE_COLOR_BLACK) {
					w->left->color = NODE_COLOR_BLACK;
					w->color = NODE_COLOR_RED;
					right_rotate(tree, w);
					w = x->parent->right;
				}
				w->color = x->parent->color;
				x->parent->color = NODE_COLOR_BLACK;
				w->right->color = NODE_COLOR_BLACK;
				left_rotate(tree, x->parent);
				x = *tree;
			}
		}
		else {
			w = x->parent->left;
			if (w->color == NODE_COLOR_RED) {
				w->color = NODE_COLOR_BLACK;
				x->parent->color = NODE_COLOR_RED;
				right_rotate(tree, x->parent);
				w = x->parent->left;
			}
			if ( (w->right->color == NODE_COLOR_BLACK) && (w->left->color == NODE_COLOR_BLACK) ) {
				w->color = NODE_COLOR_RED;
				x = x->parent;
			}
			else {
				if (w->left->color == NODE_COLOR_BLACK) {
					w->right->color = NODE_COLOR_BLACK;
					w->color = NODE_COLOR_RED;
					left_rotate(tree, w);
					w = x->parent->left;
				}
				w->color = x->parent->color;
				x->parent->color = NODE_COLOR_BLACK;
				w->left->color = NODE_COLOR_BLACK;
				right_rotate(tree, x->parent);
				x = *tree;
			}
		}
	}
	x->color = NODE_COLOR_BLACK;
}

RBTNODE *tree_minimum(RBTNODE *x) {
	RBTNODE *retval = x;
	while (retval->left != NULL_NODE) {
		retval = retval->left;
	}
	return retval;
}

RBTNODE *tree_successor(RBTNODE *x) {
	RBTNODE *xx, *yy;
	xx = x;
	if (xx->right != NULL_NODE) {
		return tree_minimum(xx->right);
	}
	yy = xx->parent;
	while ( (yy != NULL_NODE) && (xx == yy->right) ) {
		xx = yy;
		yy = yy->parent;
	}
	return yy;
}

void rbtree_insert_fixup(RBTNODE **tree, RBTNODE *z) {
	RBTNODE *y = NULL_NODE;
	//printf("BEFORE FIXUP: fixing up %ld, whose father is %ld\n", (z->key_func)(z), (z->parent->key_func)(z->parent));
	while (z->parent->color == NODE_COLOR_RED) {
		//if z's parent is red
		if (z->parent == z->parent->parent->left) {
			//if z's parent is a left child
			y = z->parent->parent->right; //y = z's uncle
			if (y->color == NODE_COLOR_RED) {
				z->parent->color = NODE_COLOR_BLACK;
				y->color = NODE_COLOR_BLACK;
				z->parent->parent->color = NODE_COLOR_RED; //change grapdparent to red (maybe root changes color?)
				z= z->parent->parent; //set z to it's grandparent - if z==root the loop will break and root will be setto black
			}
			else {
				//if z(red)'s uncle is black (e.g. if it is NULL_NODE)
				if (z == z->parent->right)  {
					//if z is a right child
					//left rotate z's parent (and set z=z->parent)
					z = z->parent;
					left_rotate(tree, z);
				}
				//right rotate current z's grandparent (z may be the original z or z->parent)
				z->parent->color = NODE_COLOR_BLACK;
				z->parent->parent->color = NODE_COLOR_RED; //some violations may take place here, e.g. root colored red (which will be taken care of with rioght rotation)
				right_rotate(tree, z->parent->parent);
				//printf("AFTER RIGHT ROTATE ROOT IS: %d\n", (*tree)->rcpt->timestamp);
			}
		}
		else {
			y = z->parent->parent->left;
			if (y->color == NODE_COLOR_RED) {
				z->parent->color = NODE_COLOR_BLACK;
				y->color = NODE_COLOR_BLACK;
				z->parent->parent->color = NODE_COLOR_RED;
				z = z->parent->parent;
			}
			else {
				if (z == z->parent->left)  {
					z = z->parent;
					right_rotate(tree, z);
				}
				z->parent->color = NODE_COLOR_BLACK;
				z->parent->parent->color = NODE_COLOR_RED;
				left_rotate(tree, z->parent->parent);
			}
		}
	}
	(*tree)->color = NODE_COLOR_BLACK;
}

void rbtree_insert(RBTNODE **tree, RBTNODE *z) {
	//printf("inserting new node: key=%ld\n", (unsigned long)z->key_func(z) );

	RBTNODE *y = NULL_NODE;
	RBTNODE *x = *tree;
	//start from root
	//insert new node as a leaf of the tree (normal binary tree insert operation)
	while (x != NULL_NODE) {
		//while x != NULL (while we haven't reached the leaf under which z will be "hanged") move pointers downwards
		//in the end, y-->z's parent
		y = x;
		if ( (z->comp_func)(z, x) == -1) {
			x = x->left;
		}
		else {
			x = x->right;
		}
	}

	//set z's father
	z->parent = y;
	//printf("INSERT NODE: setting new node (%ld) 's father to %ld\n", (unsigned long)(z->key_func)(z), (unsigned long)(y->key_func)(y));
	if (y == NULL_NODE) {
		//if y is null, this means that the tree was empty ===> z is new root
		*tree = z;
	}
	else
//	if (z->rcpt->timestamp < y->rcpt->timestamp) {
	if ( (z->comp_func)(z,  y) == -1) {
		//decide if z will be hanged left or right
		y->left = z;
	}
	else {
		y->right = z;
	}

	//set z's children pointers to NULL_NODE
	z->left = NULL_NODE;
	z->right = NULL_NODE;
	//set z's color to red
	z->color = NODE_COLOR_RED;

	//call fixup routine to rebalance tree
	rbtree_insert_fixup(tree, z);
}

RBTNODE *rbtree_delete(RBTNODE **tree, RBTNODE *z) {
	RBTNODE *x = NULL_NODE;
	RBTNODE *y = NULL_NODE;
	void *data_to_be_deleted = NULL;

	//if z has no children or exactly one child
	if ( (z->left == NULL_NODE) || (z->right == NULL_NODE) ) {
		y = z;
	}
	else {
		//if z has two children, search for the next node in sorted order
		y = tree_successor(z);
		//printf("y = z's successor is %ld\n", (unsigned long)y->key_func(y));
	}

	if (y->left != NULL_NODE) {
		//if the node to be removed has a left child
		x = y->left;
	}
	else {
		//if the node to be removed doesn't have a left child
		x = y->right;
	}

	//y is to be disconnected, y's parent will now be y's child's father
	//!!No check is performed to test whether the node to be removed had no children. In such a case, x->parent==NULL_NODE
	//things will be taken care of by rbtree_delete_fixup
	x->parent = y->parent;

	//if y is root, then it's child will be the new root
	if (y->parent == NULL_NODE) {
		*tree = x;
		//printf("x is now root: %d - %d\n", x->rcpt->timestamp, (*tree)->rcpt->timestamp);
	}
	else
	if (y == y->parent->left) {
		//if y is a left child x will become a left child too
		y->parent->left = x;
	}
	else {
		//else x ===> right child
		y->parent->right = x;
	}

	if (y != z) {
		//printf("z has 2 children\n");
		//printf("z->left = %ld\n", (unsigned long)z->left->key_func(z->left));
		//printf("z->right= %ld\n", (unsigned long)z->right->key_func(z->right));
		//if z had two children
		//y is the next node in sorted order
		//y is already disconnected and now we should reconnect it in z's place
		//z->rcpt->timestamp = y->rcpt->timestamp;
		z->set_key_func(z, y->key_func(y));
		//swap data so that y's data will be eventually freed and z->data will not be relocated
		data_to_be_deleted = z->data;
		z->data = y->data;
		y->data = data_to_be_deleted;
	}

	if (y->color == NODE_COLOR_BLACK) {
		rbtree_delete_fixup(tree, x);
	}

	return y;
}

void walk_inorder(RBTNODE *tree) {
	if (tree != NULL_NODE) {
		walk_inorder(tree->left);
		//printf("Recursion: %ld\n", (unsigned long)tree->key_func(tree));
		walk_inorder(tree->right);
	}

}

//RBTNODE *find_node(RBTNODE *tree, void *key) {
RBTNODE *rbtree_find_node(RBTNODE *tree, RBTNODE *node) {
	if (tree != NULL_NODE) {
		if ( (*tree->comp_func)(tree, node) == -1) {
			return rbtree_find_node(tree->right, node);
		}
		else
		if ( (*tree->comp_func)(tree, node) == 1) {
			return rbtree_find_node(tree->left, node);
		}
		else
		return tree;
	}
	else {
		return NULL_NODE;
	}
}

//moufa, alla O(logn)
void do_stuff(RBTNODE *tree, void *args) {
	//printf("Doing stuff...%d\n", (unsigned int)(*tree->key_func)(tree) );
}

//don't use this....
void slow_range_query(RBTNODE* tree, void (*do_stuff)(RBTNODE*, void *), void *do_stuff_args, RBTNODE *node) {
	if (tree != NULL_NODE) {
		slow_range_query(tree->left, do_stuff, do_stuff_args, node);
		if ( (*tree->comp_func)(tree, node) != -1) {
			do_stuff( tree, do_stuff_args );
		}
		slow_range_query(tree->right, do_stuff, do_stuff_args, node);
	}
}

//...use this
void rbtree_range_query(RBTNODE* tree, void (*do_stuff)(RBTNODE*, void *), void *do_stuff_args, RBTNODE *node) {
	if (tree == NULL_NODE)
		return;
	if ( (*tree->comp_func)(tree, node) != -1) {
		do_stuff(tree, do_stuff_args);
  		rbtree_range_query(tree->left, do_stuff, do_stuff_args, node);
	}
	rbtree_range_query(tree->right, do_stuff, do_stuff_args, node);
}

void print_tree(RBTNODE *tree) {
	if (tree == NULL_NODE) {
		return;
	}
	fprintf(stderr, "\t%ld(%ld / %s)\t", (unsigned long)(tree->key_func)(tree), (unsigned long)(tree->parent->key_func)(tree->parent), ((tree->color==NODE_COLOR_BLACK)?"B":"R"));
	print_tree(tree->right);
	fprintf(stderr, "\n");
	print_tree(tree->left);
}

RBTNODE *rbtree_find_min(RBTNODE *tree) {
	RBTNODE *cur = tree;
	while (cur->left != NULL_NODE) {
		cur = cur->left;
	}
	return cur;
}
